/**************************************************************
* M Display Library for the Moacon
* Michael Franklin
* 2015-02-03 - Fix for MD_CreateLayer
* 2013-09-09 - First Release
*
* To use simply add mdisplay.c and mdisplay.h to the Moacon
* Studio project and #include "mdisplay.h" in your program's 
* source file
***************************************************************/

#include "moacon500.h"
#include <stdlib.h>
#include <math.h>
#include <string.h>

#include "mdisplay.h"

#define DLE 0x10
#define STX 0x02
#define ETX 0x03

static void MD_DefaultOnError(MD_Status status)
{
	printf("M Display Error: 0x%04X\r\n", status);
}

static MD_Protocol    m_protocol              = MD_Protocol_UTF8;
static unsigned char  m_serialPort            = 0;
static unsigned short m_timeoutInMilliseconds = 1000;
static unsigned short m_checksum              = 0;
static unsigned char  m_isChecksumEnabled     = 0;

static void (*m_onError)(MD_Status) = MD_DefaultOnError;

static const double PowersOf10[] = 
{
	1, 
	10, 
	100, 
	1000, 
	10000, 
	100000, 
	1000000,
   10000000, 
	100000000, 
	1000000000
};

/*************************************************************
* Reverses a string
*************************************************************/
static void ReverseString(char* begin, char* end)
{
    char aux;
    while (end > begin)
	 {
        aux = *end, *end-- = *begin, *begin++ = aux;
	 }
}

/*************************************************************
* Converts a double to a base-10 string
*
* Copyright 2005, 2006, 2007
* Nick Galbreath -- nickg [at] modp [dot] com
* All rights reserved.
* https://code.google.com/p/stringencoders/
*************************************************************/
static unsigned int DoubleToString(char* str, double value)
{
	 int prec = 2;
	
    // Hacky test for NaN
    // under -fast-math this won't work, but then you also won't
    // have correct nan values anyways.  The alternative is
    // to link with libmath (bad) or hack IEEE double bits (bad)
    if (! (value == value)) 
	{
        str[0] = 'n'; str[1] = 'a'; str[2] = 'n'; str[3] = '\0';
        return (size_t)3;
    }
	
    // if input is larger than thres_max, revert to exponential 
    const double thres_max = (double)(0x7FFFFFFF);

    double diff = 0.0;
    char* wstr = str;

    if (prec < 0) 
	{
        prec = 0;
    } else if (prec > 9) 
	{
        // precision of >= 10 can lead to overflow errors 
        prec = 9;
    }


    // we'll work in positive values and deal with the
    //   negative sign issue later 
    int neg = 0;
    if (value < 0) {
        neg = 1;
        value = -value;
    }


    int whole = (int) value;
    double tmp = (value - whole) * PowersOf10[prec];
    u32 frac = (u32)(tmp);
    diff = tmp - frac;

    if (diff > 0.5) 
	{
        ++frac;
		
        // handle rollover, e.g.  case 0.99 with prec 1 is 1.0  
        if (frac >= PowersOf10[prec]) 
		{
            frac = 0;
            ++whole;
        }
    } else if (diff == 0.5 && ((frac == 0) || (frac & 1))) 
	{
        // if halfway, round up if odd, OR
        //  if last digit is 0.  That last part is strange 
        ++frac;
    }
	 
    if (prec == 0) 
	{
        diff = value - whole;
        if (diff > 0.5) 
		{
            // greater than 0.5, round up, e.g. 1.6 -> 2 
            ++whole;
        } else if (diff == 0.5 && (whole & 1)) 
		{
            // exactly 0.5 and ODD, then round up 
            // 1.5 -> 2, but 2.5 -> 2 
            ++whole;
        }
    } 
	else 
	{
        int count = prec;
		
        // now do fractional part, as an unsigned number
        do 
		{
            --count;
            *wstr++ = (char)(48 + (frac % 10));
        } while (frac /= 10);
		
        // add extra 0s
        while (count-- > 0) 
		{
			*wstr++ = '0';
		}
		
        // add decimal
        *wstr++ = '.';
    }

    // do whole part
    // Take care of sign
    // Conversion. Number is reversed.
    do 
	{
		*wstr++ = (char)(48 + (whole % 10)); 
	} while (whole /= 10);
	
    if (neg) 
	{
        *wstr++ = '-';
    }
    *wstr='\0';
    ReverseString(str, wstr-1);
	
    return (size_t)(wstr - str);
}

static char IntToHexChar(unsigned int i)
{
	return i <= 9 ? '0' + i : 'A' - 10 + i;
}


void MD_SetOnError(void (*onError)(MD_Status status))
{
	m_onError = onError;
}

void MD_SetProtocol(MD_Protocol protocol)
{
	m_protocol = protocol;
}

MD_Protocol MD_GetProtocol()
{
	return m_protocol;
}

unsigned char MD_IsChecksumEnabled()
{
	return m_isChecksumEnabled;
}

void MD_EnableChecksum()
{
	m_isChecksumEnabled = 1;
}

void MD_DisableChecksum()
{
	m_isChecksumEnabled = 0;
}

void MD_OpenCom(u8 comCh, u32 comBaud, u8 comMode)
{
	openCom(comCh, comBaud, comMode);
	m_serialPort = comCh;
	
	comFlush(m_serialPort);
}

void MD_SetTimeout(unsigned short timeoutInMilliseconds)
{
	m_timeoutInMilliseconds = timeoutInMilliseconds;
}

unsigned short MD_GetTimeout()
{
	return m_timeoutInMilliseconds;
}

/*************************************************************
* Sends a single byte to the M Display
*************************************************************/
static void MD_SendByte(unsigned char byte)
{
	//If binary protocol, DLE characters must be escaped
	if (m_protocol == MD_Protocol_Binary && byte == DLE)
	{
		comPut(m_serialPort, DLE);
		
		if (m_isChecksumEnabled)
		{
			m_checksum += DLE;
		}
	}
	
	comPut(m_serialPort, byte);
	
	if (m_isChecksumEnabled)
	{
		m_checksum += byte;
	}
}

/*************************************************************
* Sends test to the M Display
*************************************************************/
static void MD_SendString(const char* text)
{
	while(*text != '\0')
	{
		MD_SendByte(*text);
		text++;
	}
}

static void MD_GetStartFrame(int* time)
{	
	short c = '\0';
	
	if (m_protocol == MD_Protocol_Binary)
	{		
		int escapeNextChar = 0;
		
		//Repeatedly read bytes until tmeout
		//or start of frame '<DLE><STX>'
		do
		{
			c = comGet(m_serialPort);
			if (c != -1)
			{
				if(escapeNextChar)
				{
					if (c == STX)
					{
						break;
					}
					escapeNextChar = 0;
				}
				else if (c == DLE)
				{
					escapeNextChar = 1;
				}
			}
			(*time)--;
		}	while((*time) >= 0);
	}
	else if (m_protocol == MD_Protocol_UTF8)
	{
		//Repeatedly read bytes until tmeout
		//or start of frame '\n'
		do
		{
			c = comGet(m_serialPort);
			(*time)--;
		}	while(c != '\n' && (*time) >= 0);
	}
}

static int MD_GetUntilEndFrame(char* buffer, unsigned int bufferSize, int* time)
{
	short c = '\0';
	int i = 0;
	

	if (m_protocol == MD_Protocol_Binary)
	{		
		int escapeNextChar = 0;
		
		//Repeatedly read bytes until tmeout
		//or end of frame '<DLE><ETX>'
		do
		{
			c = comGet(m_serialPort);
			if (c != -1)
			{
				if(escapeNextChar)
				{
					if (c == ETX)
					{
						break;
					}
					buffer[i] = c;
					i++;
					escapeNextChar = 0;
				}
				else if (c == DLE)
				{
					escapeNextChar = 1;
				}
				else
				{
					buffer[i] = c;
					i++;
				}
			}
			(*time)--;
		} while((*time) >= 0 && i < bufferSize);
		
	}
	else if (m_protocol == MD_Protocol_UTF8)
	{
		//Repeatedly read bytes until tmeout
		//or end of frame '\r'
		do
		{
			c = comGet(m_serialPort);
			if (c != -1)
			{
				buffer[i] = c;
				i++;
			}
			(*time)--;
		}	while(c != '\r' && (*time) >= 0 && i < bufferSize);
	}
	
	return i;
}

static unsigned char IsAckOrNak(const unsigned char* buffer)
{
	if (m_protocol == MD_Protocol_Binary)
	{
		if (*buffer != 0x06 && *buffer != 0x15)
		{
			//negative response
			return 0;
		}
	}
	else if (m_protocol == MD_Protocol_UTF8)
	{
		if (strncmp(buffer, "Ack", 3) != 0 && strncmp(buffer, "Nak", 3) != 0)
		{
			//negative response
			return 0;
		}
	}
	
	//Positive response
	return 1;
}

static MD_Status GetResponseCode(unsigned char* buffer, int bufferSize)
{
	MD_Status status = MD_Status_InvalidResponse;
	
	if (m_protocol == MD_Protocol_Binary)
	{
		//skip Ack or Nak
		buffer++; 
		bufferSize -= 1;
		
		//If we have enough data in the buffer for a response code
		if (bufferSize >= 2)
		{
			status = *((const unsigned short*)buffer);
		}
	}
	else if (m_protocol == MD_Protocol_UTF8)
	{
		//skip Ack or Nak plus space
		buffer += 4;
		bufferSize -= 4;
		
		//If we have enough data in the buffer for a response code
		if (bufferSize >= 4)
		{
			bufferSize--;                  //point to ending '\r'
			buffer[bufferSize] = '\0';     //Add null terminating character, overwriting '\r'
			
			//Convert last 4 characters from hex string to integer response code
			status = (int)strtol(buffer, NULL, 16);
		}
	}
	
	return status;
}

/*************************************************************
* Reads response from the M Display
*************************************************************/
static MD_Status MD_ProcessResponse()
{
	short c;
	char buffer[32];
	int i = 0;
	
	int time = m_timeoutInMilliseconds * 1000000 / 931;  //This formula was determined empircally
	
	//Get start of frame
	MD_GetStartFrame(&time);
	
	//If timed out
	if (time < 0)
	{
		m_onError(MD_Status_Timeout);
		return MD_Status_Timeout;
	}
	else
	{
		//Get end of frame
		i = MD_GetUntilEndFrame(buffer, 32, &time);
	}
	
	//If timed out
	if (time < 0)
	{
		m_onError(MD_Status_Timeout);
		return MD_Status_Timeout;
	}
	
	//If not a response
	if (!IsAckOrNak(buffer))
	{
		printf("No ack\r\n");
		m_onError(MD_Status_InvalidResponse);
		return MD_Status_InvalidResponse;
	}
	
	MD_Status responseCode = GetResponseCode(buffer, i);
	
	//If not a successful response, call error handler
	if (responseCode != MD_Status_Success)
	{
		m_onError(responseCode);
	}
	
	//I don't know why this is necessary, but without it
	//the Moacon comes to a halt
	delay(1);
	
	return responseCode;
}

/*************************************************************
* Sends a start of frame to the M Display
*************************************************************/
static void MD_StartFrame()
{
	comFlush(m_serialPort);
	
	//Reset checksum accumulation
	if (m_isChecksumEnabled)
	{
		m_checksum = 0;
	}
	
	if (m_protocol == MD_Protocol_Binary)
	{
		comPut(m_serialPort, DLE);
		comPut(m_serialPort, STX);
	}
	else if (m_protocol == MD_Protocol_UTF8)
	{
		comPut(m_serialPort, '\n');
	}
}

/*************************************************************
* Sends the start of a command argument
*************************************************************/
static void MD_ArgStart()
{
	//Arguments are delimitted by a space for UTF-8
	if (m_protocol == MD_Protocol_UTF8)
	{
		MD_SendString(" ");
	}
}

/*************************************************************
* Send checksum argument
*************************************************************/
static void MD_ArgChecksum()
{
	MD_ArgStart();
	
	if (m_protocol == MD_Protocol_Binary)
	{
		//Create a local copy of m_checksum
		//So MD_SendByte doesn't change m_checksum on us
		unsigned short checksum = m_checksum;
		unsigned char* byte = (unsigned char*)&checksum;
		
		MD_SendByte(byte[0]);
		MD_SendByte(byte[1]);
	}
	else if (m_protocol == MD_Protocol_UTF8)
	{
		int i = 0;
		for(i = 3; i >= 0; i--)
		{
			comPut(m_serialPort, IntToHexChar((m_checksum >> (i * 4)) & 0xF));
		}
	}
}

/*************************************************************
* Sends an end of frame to the M Display
*************************************************************/
static void MD_EndFrame()
{
	//Append checksum as last argument
	if (m_isChecksumEnabled)
	{
		MD_ArgChecksum();
	}
		
	if (m_protocol == MD_Protocol_Binary)
	{
		comPut(m_serialPort, DLE);
		comPut(m_serialPort, ETX);
	}
	else if (m_protocol == MD_Protocol_UTF8)
	{
		comPut(m_serialPort, '\r');
	}
}

/*************************************************************
* Sends the command to the M Display
*************************************************************/
static void MD_SendCommand(MD_Command command)
{
	if (m_protocol == MD_Protocol_Binary)
	{
		MD_SendByte(*(unsigned char*)&command);
	}
	else if (m_protocol == MD_Protocol_UTF8)
	{
		switch(command)
		{
			case MD_Command_ArcTo:
				MD_SendString("ArcTo");
				break;
				
			case MD_Command_Backlight:
				MD_SendString("Backlight");
				break;
				
			case MD_Command_BevelButton:
				MD_SendString("BevelButton");
				break;
				
			case MD_Command_Button:
				MD_SendString("Button");
				break;
				
			case MD_Command_Clear:
				MD_SendString("Clear");
				break;
				
			case MD_Command_ClearPath:
				MD_SendString("ClearPath");
				break;
				
			case MD_Command_Color:
				MD_SendString("Color");
				break;
				
			case MD_Command_CreateLayer:
				MD_SendString("CreateLayer");
				break;
				
			case MD_Command_CurveTo:
				MD_SendString("CurveTo");
				break;
				
			case MD_Command_Delay:
				MD_SendString("Delay");
				break;
				
			case MD_Command_DestroyLayer:
				MD_SendString("DestroyLayer");
				break;
				
			case MD_Command_DisableFlush:
				MD_SendString("DisableFlush");
				break;
				
			case MD_Command_DrawArc:
				MD_SendString("DrawArc");
				break;
				
			case MD_Command_DrawCircle:
				MD_SendString("DrawCircle");
				break;
				
			case MD_Command_DrawCurve:
				MD_SendString("DrawCurve");
				break;
				
			case MD_Command_DrawEllipse:
				MD_SendString("DrawEllipse");
				break;
				
			case MD_Command_DrawLine:
				MD_SendString("DrawLine");
				break;
				
			case MD_Command_DrawPath:
				MD_SendString("DrawPath");
				break;
				
			case MD_Command_DrawRectangle:
				MD_SendString("DrawRectangle");
				break;
				
			case MD_Command_DrawText:
				MD_SendString("DrawText");
				break;
				
			case MD_Command_FillCircle:
				MD_SendString("FillCircle");
				break;
				
			case MD_Command_FillEllipse:
				MD_SendString("FillEllipse");
				break;
				
			case MD_Command_FillPath:
				MD_SendString("FillPath");
				break;
				
			case MD_Command_FillRectangle:
				MD_SendString("FillRectangle");
				break;
				
			case MD_Command_FillText:
				MD_SendString("FillText");
				break;
				
			case MD_Command_FontSize:
				MD_SendString("FontSize");
				break;
				
			case MD_Command_GlossyButton:
				MD_SendString("GlossyButton");
				break;
				
			case MD_Command_Image:
				MD_SendString("Image");
				break;
				
			case MD_Command_Layer:
				MD_SendString("Layer");
				break;
				
			case MD_Command_LineCap:
				MD_SendString("LineCap");
				break;
				
			case MD_Command_LineJoin:
				MD_SendString("LineJoin");
				break;
				
			case MD_Command_LineTo:
				MD_SendString("LineTo");
				break;
				
			case MD_Command_LineWidth:
				MD_SendString("LineWidth");
				break;
				
			case MD_Command_LinearGradient:
				MD_SendString("LinearGradient");
				break;
				
			case MD_Command_LoadFont:
				MD_SendString("LoadFont");
				break;
				
			case MD_Command_MoveLayer:
				MD_SendString("MoveLayer");
				break;
				
			case MD_Command_MoveTo:
				MD_SendString("MoveTo");
				break;
				
			case MD_Command_Ping:
				MD_SendString("Ping");
				break;
				
			case MD_Command_RadialGradient:
				MD_SendString("RadialGradient");
				break;
				
			case MD_Command_RunScript:
				MD_SendString("RunScript");
				break;
				
			case MD_Command_Reset:
				MD_SendString("Reset");
				break;
				
			case MD_Command_SetPixel:
				MD_SendString("SetPixel");
				break;
				
			case MD_Command_UnloadFont:
				MD_SendString("UnloadFont");
				break;
				
			case MD_Command_ClosePath:
				MD_SendString("ClosePath");
				break;
				
			case MD_Command_DisableLayer:
				MD_SendString("DisableLayer");
				break;
				
			case MD_Command_EnableFlush:
				MD_SendString("EnableFlush");
				break;
				
			case MD_Command_EnableLayer:
				MD_SendString("EnableLayer");
				break;
		}
		
	}
}

/*************************************************************
* Sends an unsigned 8-bit integer argument to the M Display
*************************************************************/
static void MD_ArgUInt8(unsigned char value)
{
	MD_ArgStart();
	
	if (m_protocol == MD_Protocol_Binary)
	{
		MD_SendByte(value);
	}
	else if (m_protocol == MD_Protocol_UTF8)
	{
		char buffer[16];
		sprintf(buffer, "%d", (int)value);
		
		MD_SendString(buffer);
	}
}

/*************************************************************
* Sends a signed 16-bit integer argument to the M Display
*************************************************************/
static void MD_ArgInt16(short value)
{
	MD_ArgStart();
	
	if (m_protocol == MD_Protocol_Binary)
	{
		unsigned char* byte = (unsigned char*)&value;
		
		MD_SendByte(byte[0]);
		MD_SendByte(byte[1]);
	}
	else if (m_protocol == MD_Protocol_UTF8)
	{
		char buffer[16];
		sprintf(buffer, "%d", (int)value);
	
		MD_SendString(buffer);
	}
}

/*************************************************************
* Sends an unsigned 16-bit integer argument to the M Display
*************************************************************/
static void MD_ArgUInt16(unsigned short value)
{
	MD_ArgStart();
	
	if (m_protocol == MD_Protocol_Binary)
	{
		unsigned char* byte = (unsigned char*)&value;
		
		MD_SendByte(byte[0]);
		MD_SendByte(byte[1]);
	}
	else if (m_protocol == MD_Protocol_UTF8)
	{
		char buffer[16];
		sprintf(buffer, "%d", (int)value);
	
		MD_SendString(buffer);
	}
}

/*************************************************************
* Sends an unsigned 32-bit integer argument to the M Display
*************************************************************/
static void MD_ArgUInt32(unsigned int value)
{
	MD_ArgStart();
	
	if (m_protocol == MD_Protocol_Binary)
	{
		unsigned char* byte = (unsigned char*)&value;
		
		MD_SendByte(byte[0]);
		MD_SendByte(byte[1]);
		MD_SendByte(byte[2]);
		MD_SendByte(byte[3]);
	}
	else if (m_protocol == MD_Protocol_UTF8)
	{
		char buffer[16];
		sprintf(buffer, "%d", (int)value);
		
		MD_SendString(buffer);
	}
}

/*************************************************************
* Sends a real number argument to the M Display
*************************************************************/
static void MD_ArgReal(double value)
{
	MD_ArgStart();
	
	if (m_protocol == MD_Protocol_Binary)
	{
		//Real numbers are sent to M Display multiplied 
		//by 10 so they can be sent as 16-bit integers
		short roundedValue = round(value) * 10;
		
		unsigned char* byte = (unsigned char*)&roundedValue;
		
		MD_SendByte(byte[0]);
		MD_SendByte(byte[1]);
	}
	else if (m_protocol == MD_Protocol_UTF8)
	{
		char buffer[20];
		memset(buffer, 0, 20);
		
		DoubleToString(buffer, value);
		MD_SendString(buffer);
	}
}

/*************************************************************
* Sends a color argument to the M Display
*************************************************************/
static void MD_ArgColor(unsigned int color)
{
	MD_ArgStart();
	
	if (m_protocol == MD_Protocol_Binary)
	{
		unsigned char* byte = (unsigned char*)&color;
		
		MD_SendByte(byte[0]);
		MD_SendByte(byte[1]);
		MD_SendByte(byte[2]);
		MD_SendByte(byte[3]);
	}
	else if (m_protocol == MD_Protocol_UTF8)
	{
		int i = 0;
		for(i = 7; i >= 0; i--)
		{
			MD_SendByte(IntToHexChar((color >> (i * 4)) & 0xF));
		}
	}
}

/*************************************************************
* Sends a text argument to the M Display
*************************************************************/
static void MD_ArgText(const char* text)
{
	MD_ArgStart();
	
	if (m_protocol == MD_Protocol_Binary)
	{
		MD_SendString(text);
		
		//Add null terminating character
		MD_SendByte('\0');
	}
	else if (m_protocol == MD_Protocol_UTF8)
	{
		//surround argument in quotes
		MD_SendString("\"");
		
		while(*text != '\0')
		{
			//escape double quotes and backslashes
			if (*text == '\"' || *text == '\\')
			{
				MD_SendByte('\\');
			}
			MD_SendByte(*text);
			text++;
		}
		
		//surround argument in quotes
		MD_SendString("\"");
	}
}

/*************************************************************
* Sends a line cap argument to the M Display
*************************************************************/
static void MD_ArgLineCap(MD_Cap lineCap)
{
	if (m_protocol == MD_Protocol_Binary)
	{
		MD_ArgUInt8((unsigned char)lineCap);
	}
	else if (m_protocol == MD_Protocol_UTF8)
	{		
		switch(lineCap)
		{
			case MD_Cap_Butt:
				MD_ArgText("Butt");
				break;
			
			case MD_Cap_Round:
				MD_ArgText("Round");
				break;
			
			case MD_Cap_Square:
				MD_ArgText("Square");
				break;
				
		}
	}
}

/*************************************************************
* Sends a line join to the M Display
*************************************************************/
static void MD_ArgLineJoin(MD_Join lineJoin)
{
	if (m_protocol == MD_Protocol_Binary)
	{
		MD_ArgUInt8((unsigned char)lineJoin);
	}
	else if (m_protocol == MD_Protocol_UTF8)
	{
		switch(lineJoin)
		{
			case MD_Join_Miter:
				MD_ArgText("Miter");
				break;
			
			case MD_Join_Round:
				MD_ArgText("Round");
				break;
			
			case MD_Join_Bevel:
				MD_ArgText("Bevel");
				break;
				
		}
	}
}

/*************************************************************
* Send only command
*************************************************************/
static MD_Status MD_Send_c(MD_Command command)
{
	MD_StartFrame();
	
	MD_SendCommand(command);
	
	MD_EndFrame();
	
	return MD_ProcessResponse();
}

/*************************************************************
* Send command and text
*************************************************************/
static MD_Status MD_Send_ct(MD_Command command, const char* a0)
{
	MD_StartFrame();
	
	MD_SendCommand(command);
	
	MD_ArgText(a0);
	
	MD_EndFrame();
	
	return MD_ProcessResponse();
}

/*************************************************************
* Send command, 2 16-bit signed integers, and text
*************************************************************/
static MD_Status MD_Send_ci162t(MD_Command command, short a0, short a1, const char* a2)
{
	MD_StartFrame();
	
	MD_SendCommand(command);
	
	MD_ArgInt16(a0);
	MD_ArgInt16(a1);
	MD_ArgText(a2);
	
	MD_EndFrame();
	
	return MD_ProcessResponse();
}

/*************************************************************
* Send command, 2 16-bit signed integers, and 2 16-bit unsigned 
* integers
*************************************************************/
static MD_Status MD_Send_ci162u162(MD_Command command, short a0, short a1, unsigned short a2, unsigned short a3)
{
	MD_StartFrame();
	
	MD_SendCommand(command);
	
	MD_ArgInt16(a0);
	MD_ArgInt16(a1);
	MD_ArgUInt16(a2);
	MD_ArgUInt16(a3);
	
	MD_EndFrame();
	
	return MD_ProcessResponse();
}

/*************************************************************
* Send command, 2 16-bit signed integers, 2 16-bit unsigned 
* integers, and 1 unsigned byte
*************************************************************/
static MD_Status MD_Send_ci162u162u8(MD_Command command, short a0, short a1, unsigned short a2, unsigned short a3, unsigned char a4)
{
	MD_StartFrame();
	
	MD_SendCommand(command);
	
	MD_ArgInt16(a0);
	MD_ArgInt16(a1);
	MD_ArgUInt16(a2);
	MD_ArgUInt16(a3);
	
	MD_ArgUInt8(a4);
	
	MD_EndFrame();
	
	return MD_ProcessResponse();
}

/*************************************************************
* Send command and color
*************************************************************/
static MD_Status MD_Send_cc(MD_Command command, unsigned int color)
{
	MD_StartFrame();
	
	MD_SendCommand(command);
	
	MD_ArgColor(color);
	
	MD_EndFrame();
	
	return MD_ProcessResponse();
}

/*************************************************************
* send command and an unsigned 8-bit integer
*************************************************************/
static MD_Status MD_Send_cu8(MD_Command command, unsigned char value)
{
	MD_StartFrame();
	
	MD_SendCommand(command);
	
	MD_ArgUInt8(value);
	
	MD_EndFrame();
	
	return MD_ProcessResponse();
}

/*************************************************************
* send command and an unsigned 32-bit integer
*************************************************************/
static MD_Status MD_Send_cu32(MD_Command command, unsigned int value)
{
	MD_StartFrame();
	
	MD_SendCommand(command);
	
	MD_ArgUInt32(value);
	
	MD_EndFrame();
	
	return MD_ProcessResponse();
}

/*************************************************************
* send command and 1 real number
*************************************************************/
static MD_Status MD_Send_cd(MD_Command command, double a0)
{
	MD_StartFrame();
	
	MD_SendCommand(command);
	
	MD_ArgReal(a0);
	
	MD_EndFrame();
	
	return MD_ProcessResponse();
}

/*************************************************************
* send command and 1 real number
*************************************************************/
static MD_Status MD_Send_cd2(MD_Command command, double a0, double a1)
{
	MD_StartFrame();
	
	MD_SendCommand(command);
	
	MD_ArgReal(a0);
	MD_ArgReal(a1);
	
	MD_EndFrame();
	
	return MD_ProcessResponse();
}

/*************************************************************
* send command and 3 real numbers
*************************************************************/
static MD_Status MD_Send_cd3(MD_Command command, double a0, double a1, double a2)
{
	MD_StartFrame();
	
	MD_SendCommand(command);
	
	MD_ArgReal(a0);
	MD_ArgReal(a1);
	MD_ArgReal(a2);
	
	MD_EndFrame();
	
	return MD_ProcessResponse();
}

/*************************************************************
* send command and 4 real numbers
*************************************************************/
static MD_Status MD_Send_cd4(MD_Command command, double a0, double a1, double a2, double a3)
{
	MD_StartFrame();
	
	MD_SendCommand(command);
	
	MD_ArgReal(a0);
	MD_ArgReal(a1);
	MD_ArgReal(a2);
	MD_ArgReal(a3);
	
	MD_EndFrame();
	
	return MD_ProcessResponse();
}

/*************************************************************
* send command, 4 real numbers, and 3 colors
*************************************************************/
static MD_Status MD_Send_cd4c3(
	MD_Command command, 
	double a0, double a1, double a2, double a3,
	unsigned int a4, unsigned int a5, unsigned int a6)
{
	MD_StartFrame();
	
	MD_SendCommand(command);
	
	MD_ArgReal(a0);
	MD_ArgReal(a1);
	MD_ArgReal(a2);
	MD_ArgReal(a3);
	
	MD_ArgColor(a4);
	MD_ArgColor(a5);
	MD_ArgColor(a6);
	
	MD_EndFrame();
	
	return MD_ProcessResponse();
}

/*************************************************************
* send command, 4 real numbers, 3 colors, and text
*************************************************************/
static MD_Status MD_Send_cd4c3t(
	MD_Command command, 
	double a0, double a1, double a2, double a3,
	unsigned int a4, unsigned int a5, unsigned int a6,
	const char* a7)
{
	MD_StartFrame();
	
	MD_SendCommand(command);
	
	MD_ArgReal(a0);
	MD_ArgReal(a1);
	MD_ArgReal(a2);
	MD_ArgReal(a3);
	
	MD_ArgColor(a4);
	MD_ArgColor(a5);
	MD_ArgColor(a6);
	
	MD_ArgText(a7);
	
	MD_EndFrame();
	
	return MD_ProcessResponse();
}

/*************************************************************
* send command and 5 real numbers
*************************************************************/
static MD_Status MD_Send_cd5(MD_Command command, double a0, double a1, double a2, double a3, double a4)
{
	MD_StartFrame();
	
	MD_SendCommand(command);
	
	MD_ArgReal(a0);
	MD_ArgReal(a1);
	MD_ArgReal(a2);
	MD_ArgReal(a3);
	MD_ArgReal(a4);
	
	MD_EndFrame();
	
	return MD_ProcessResponse();
}

/*************************************************************
* send command, 5 real numbers, and 2 colors
*************************************************************/
static MD_Status MD_Send_cd5c2(
	MD_Command command, 
	double a0, double a1, double a2, double a3,
	double a4, 
	unsigned int a5, unsigned int a6)
{
	MD_StartFrame();
	
	MD_SendCommand(command);
	
	MD_ArgReal(a0);
	MD_ArgReal(a1);
	MD_ArgReal(a2);
	MD_ArgReal(a3);
	
	MD_ArgReal(a4);
	
	MD_ArgColor(a5);
	MD_ArgColor(a6);
	
	MD_EndFrame();
	
	return MD_ProcessResponse();
}

/*************************************************************
* send command, 5 real numbers, 2 colors, and text
*************************************************************/
static MD_Status MD_Send_cd5c2t(
	MD_Command command, 
	double a0, double a1, double a2, double a3,
	double a4, 
	unsigned int a5, unsigned int a6,
	const char* a7)
{
	MD_StartFrame();
	
	MD_SendCommand(command);
	
	MD_ArgReal(a0);
	MD_ArgReal(a1);
	MD_ArgReal(a2);
	MD_ArgReal(a3);
	
	MD_ArgReal(a4);
	
	MD_ArgColor(a5);
	MD_ArgColor(a6);
	
	MD_ArgText(a7);
	
	MD_EndFrame();
	
	return MD_ProcessResponse();
}

/*************************************************************
* send command, 5 real numbers and an unsigned 8-bit integer
*************************************************************/
static MD_Status MD_Send_cd5u8(
	MD_Command command, 
	double a0, double a1, double a2, double a3, double a4, 
	unsigned char a5)
{
	MD_StartFrame();
	
	MD_SendCommand(command);
	
	MD_ArgReal(a0);
	MD_ArgReal(a1);
	MD_ArgReal(a2);
	MD_ArgReal(a3);
	
	MD_ArgReal(a4);
	
	MD_ArgUInt8(a5);
	
	MD_EndFrame();
	
	return MD_ProcessResponse();
}

/*************************************************************
* send command, 5 real numbers, and 2 unsigned 8-bit integers
*************************************************************/
static MD_Status MD_Send_cd5u82(
	MD_Command command, 
	double a0, double a1, double a2, double a3, double a4, 
	unsigned char a5, unsigned char a6)
{
	MD_StartFrame();
	
	MD_SendCommand(command);
	
	MD_ArgReal(a0);
	MD_ArgReal(a1);
	MD_ArgReal(a2);
	MD_ArgReal(a3);
	
	MD_ArgReal(a4);
	
	MD_ArgUInt8(a5);
	MD_ArgUInt8(a6);
	
	MD_EndFrame();
	
	return MD_ProcessResponse();
}

/*************************************************************
* send command and 6 real numbers
*************************************************************/
static MD_Status MD_Send_cd6(
	MD_Command command, 
	double a0, double a1, double a2, double a3, double a4, double a5)
{
	MD_StartFrame();
	
	MD_SendCommand(command);
	
	MD_ArgReal(a0);
	MD_ArgReal(a1);
	MD_ArgReal(a2);
	MD_ArgReal(a3);
	MD_ArgReal(a4);
	MD_ArgReal(a5);
	
	MD_EndFrame();
	
	return MD_ProcessResponse();
}

/*************************************************************
* send command and 8 real numbers
*************************************************************/
static MD_Status MD_Send_cd8(
	unsigned short command, 
	double a0, double a1, double a2, double a3,
	double a4, double a5, double a6, double a7)
{
	MD_StartFrame();
	
	MD_SendCommand(command);
	
	MD_ArgReal(a0);
	MD_ArgReal(a1);
	MD_ArgReal(a2);
	MD_ArgReal(a3);
	MD_ArgReal(a4);
	MD_ArgReal(a5);
	MD_ArgReal(a6);
	MD_ArgReal(a7);
	
	MD_EndFrame();
	
	return MD_ProcessResponse();
}

/*************************************************************
* send command, 2 real numbers, and a text
*************************************************************/
static MD_Status MD_Send_cd2t(unsigned short command, double a0, double a1, const char* a2)
{
	MD_StartFrame();
	
	MD_SendCommand(command);
	
	MD_ArgReal(a0);
	MD_ArgReal(a1);
	
	MD_ArgText(a2);
	
	MD_EndFrame();
	
	return MD_ProcessResponse();
}

/*************************************************************
* send command, 2 real numbers, text, and then 1 real number
*************************************************************/
static MD_Status MD_Send_cd2td(unsigned short command, double a0, double a1, const char* a2, double a3)
{
	MD_StartFrame();
	
	MD_SendCommand(command);
	
	MD_ArgReal(a0);
	MD_ArgReal(a1);
	
	MD_ArgText(a2);
	
	MD_ArgReal(a3);
	
	MD_EndFrame();
	
	return MD_ProcessResponse();
}

MD_Status MD_ArcTo(double x, double y, double horizontalRadius)
{
	return MD_Send_cd3(
		MD_Command_ArcTo, 
		x, y, horizontalRadius);
}

MD_Status MD_ArcTo1(double x, double y, double horizontalRadius, double verticalRadius)
{
	return MD_Send_cd4(
		MD_Command_ArcTo, 
		x, y, horizontalRadius, verticalRadius);
}

MD_Status MD_ArcTo2(double x, double y, double horizontalRadius, double verticalRadius, double rotationAngle)
{
	return MD_Send_cd5(
		MD_Command_ArcTo, 
		x, y, horizontalRadius, verticalRadius,
		rotationAngle);
}

MD_Status MD_ArcTo3(
	double x, double y, double horizontalRadius, double verticalRadius, double rotationAngle, 
	unsigned char largeArc)
{
	return MD_Send_cd5i8(
		MD_Command_ArcTo, 
		x, y, horizontalRadius, verticalRadius, 
		rotationAngle,
		largeArc);
}

MD_Status MD_ArcTo4(
	double x, double y, double horizontalRadius, double verticalRadius, double rotationAngle, 
	unsigned char largeArc, unsigned char sweep)
{
	return MD_Send_cd5i82(
		MD_Command_ArcTo, 
		x, y, horizontalRadius, verticalRadius, 
		rotationAngle,
		largeArc, sweep);
}

MD_Status MD_Backlight(unsigned char value)
{
	return MD_Send_cu8(MD_Command_Backlight, value);
}

MD_Status MD_BevelButton(
	double topLeftX, double topLeftY, double width, double height,
	unsigned int buttonColor, unsigned int topLeftColor, unsigned int bottomRightColor)
{
	return MD_Send_cd4c3(
		MD_Command_BevelButton,
		topLeftX, topLeftY, width, height,
		buttonColor, topLeftColor, bottomRightColor);
}

MD_Status MD_BevelButton1(
	double topLeftX, double topLeftY, double width, double height,
	unsigned int buttonColor, unsigned int topLeftColor, unsigned int bottomRightColor,
	const char* text)
{
	return MD_Send_cd4c3t(
		MD_Command_BevelButton,
		topLeftX, topLeftY, width, height,
		buttonColor, topLeftColor, bottomRightColor,
		text);
}

MD_Status MD_Button(
	double topLeftX, double topLeftY, double width, double height, 
	double cornerRadius,
	unsigned int buttonColor, unsigned int borderColor)
{
	return MD_Send_cd5c2(
		MD_Command_Button,
		topLeftX, topLeftY, width, height, cornerRadius,
		buttonColor, borderColor);
}
	
MD_Status MD_Button1(
	double topLeftX, double topLeftY, double width, double height, 
	double cornerRadius,
	unsigned int buttonColor, unsigned int borderColor,
	const char* text)
{
	return MD_Send_cd5c2t(
		MD_Command_Button,
		topLeftX, topLeftY, width, height, cornerRadius,
		buttonColor, borderColor,
		text);
}

MD_Status MD_Clear()
{
	return MD_Send_c(MD_Command_Clear);
}

MD_Status MD_Clear1(short topLeftX, short topLeftY, unsigned short width, unsigned short height)
{
	return MD_Send_ci162u162(
		MD_Command_Clear, topLeftX, topLeftY, width, height);
}

MD_Status MD_ClearPath()
{
	return MD_Send_c(MD_Command_ClearPath);
}

MD_Status MD_ClosePath()
{
	return MD_Send_c(MD_Command_ClosePath);
}

MD_Status MD_Color(unsigned int color)
{
	return MD_Send_cc(MD_Command_Color, color);
}

MD_Status MD_CreateLayer(short topLeftX, short topLeftY, unsigned short width, unsigned short height, unsigned char layerID)
{
	return MD_Send_ci162u162u8(MD_Command_CreateLayer, topLeftX, topLeftY, width, height, layerID);
}

MD_Status MD_CurveTo(double x, double y, double control1X, double control1Y)
{
	return MD_Send_cd4(MD_Command_CurveTo, x, y, control1X, control1Y);
}

MD_Status MD_CurveTo1(double x, double y, double control1X, double control1Y, double control2X, double control2Y)
{
	return MD_Send_cd6(MD_Command_CurveTo, x, y, control1X, control1Y, control2X, control2Y);
}

MD_Status MD_Delay(unsigned int milliseconds)
{
	return MD_Send_cu32(MD_Command_Delay, milliseconds);
}

MD_Status MD_DestroyLayer(unsigned char layerID)
{
	return MD_Send_cu8(MD_Command_DestroyLayer, layerID);
}

MD_Status MD_DisableFlush()
{
	return MD_Send_c(MD_Command_DisableFlush);
}

MD_Status MD_DisableLayer(unsigned char layerID)
{
	return MD_Send_cu8(MD_Command_DisableLayer, layerID);
}

MD_Status MD_DrawArc(
	double centerX, double centerY, 
	double horizontalRadius, 
	double startAngle, double sweep)
{
	return MD_Send_cd5(
		MD_Command_DrawArc, 
		centerX, centerY, 
		horizontalRadius, 
		startAngle, sweep);
}

MD_Status MD_DrawArc1(
	double centerX, double centerY, 
	double horizontalRadius, double verticalRadius, 
	double startAngle, double sweep)
{
	return MD_Send_cd6(
		MD_Command_DrawArc, 
		centerX, centerY, 
		horizontalRadius, verticalRadius,
		startAngle, sweep);
}

MD_Status MD_DrawCircle(double centerX, double centerY, double radius)
{
	return MD_Send_cd3(
		MD_Command_DrawCircle, 
		centerX, centerY, radius);
}

MD_Status MD_DrawCurve(
	double fromX, double fromY, 
	double control0X, double control0Y, 
	double toX, double toY)
{
	return MD_Send_cd6(
		MD_Command_DrawCurve, 
		fromX, fromY,
		control0X, control0Y,
		toX, toY);
}

MD_Status MD_DrawCurve1(
	double fromX, double fromY, 
	double control0X, double control0Y, 
	double control1X, double control1Y, 
	double toX, double toY)
{
	return MD_Send_cd8(
		MD_Command_DrawCurve, 
		fromX, fromY,
		control0X, control0Y,
		control1X, control1Y,
		toX, toY);
}

MD_Status MD_DrawEllipse(double centerX, double centerY, double horizontalRadius, double verticalRadius)
{
	return MD_Send_cd4(MD_Command_DrawEllipse, centerX, centerY, horizontalRadius, verticalRadius);
}

MD_Status MD_DrawPath()
{
	return MD_Send_c(MD_Command_DrawPath);
}

MD_Status MD_DrawLine(double x0, double y0, double x1, double y1)
{
	return MD_Send_cd4(
		MD_Command_DrawLine, 
		x0, y0, x1, y1);
}

MD_Status MD_DrawRectangle(double topLeftX, double topLeftY, double width, double height)
{	
	return MD_Send_cd4(MD_Command_DrawRectangle, topLeftX, topLeftY, width, height);
}

MD_Status MD_DrawRectangle1(
	double topLeftX, double topLeftY, double width, double height, 
	double cornerRadius)
{
	return MD_Send_cd5(
		MD_Command_DrawRectangle, 
		topLeftX, topLeftY, width, height,
		cornerRadius);
}

MD_Status MD_DrawRectangle2(
	double topLeftX, double topLeftY, double width, double height, 
	double topLeftCornerRadius, double topRightCornerRadius, 
	double bottomLeftCornerRadius, double bottomRightCornerRadius)
{
	return MD_Send_cd8(
		MD_Command_DrawRectangle, 
		topLeftX, topLeftY, width, height,
		topLeftCornerRadius, topRightCornerRadius,
		bottomLeftCornerRadius, bottomRightCornerRadius);
}

MD_Status MD_DrawText(double baselineX, double baselineY, const char* text)
{
	return MD_Send_cd2t(
		MD_Command_DrawText, 
		baselineX, baselineY, text);
}

MD_Status MD_DrawText1(double baselineX, double baselineY, const char* text, double angle)
{
	return MD_Send_cd2td(
		MD_Command_DrawText, 
		baselineX, baselineY, text, angle);
}

MD_Status MD_EnableFlush()
{
	return MD_Send_c(MD_Command_EnableFlush);
}

MD_Status MD_EnableLayer(unsigned char layerID)
{
	return MD_Send_cu8(MD_Command_EnableLayer, layerID);
}

MD_Status MD_FillCircle(double centerX, double centerY, double radius)
{
	return MD_Send_cd3(
		MD_Command_FillCircle, 
		centerX, centerY, radius);
}

MD_Status MD_FillEllipse(double centerX, double centerY, double horizontalRadius, double verticalRadius)
{
	return MD_Send_cd4(MD_Command_FillEllipse, centerX, centerY, horizontalRadius, verticalRadius);
}

MD_Status MD_FillPath()
{
	return MD_Send_c(MD_Command_FillPath);
}

MD_Status MD_FillRectangle(double topLeftX, double topLeftY, double width, double height)
{	
	return MD_Send_cd4(
		MD_Command_FillRectangle, 
		topLeftX, topLeftY, width, height);
}

MD_Status MD_FillRectangle1(
	double topLeftX, double topLeftY, double width, double height, 
	double cornerRadius)
{
	return MD_Send_cd5(
		MD_Command_FillRectangle, 
		topLeftX, topLeftY, width, height,
		cornerRadius);
}

MD_Status MD_FillRectangle2(
	double topLeftX, double topLeftY, double width, double height, 
	double topLeftCornerRadius, double topRightCornerRadius, 
	double bottomLeftCornerRadius, double bottomRightCornerRadius)
{
	return MD_Send_cd8(
		MD_Command_FillRectangle, 
		topLeftX, topLeftY, width, height,
		topLeftCornerRadius, topRightCornerRadius,
		bottomLeftCornerRadius, bottomRightCornerRadius);
}

MD_Status MD_FillText(double baselineX, double baselineY, const char* text)
{
	return MD_Send_cd2t(
		MD_Command_FillText, 
		baselineX, baselineY, text);
}

MD_Status MD_FillText1(double baselineX, double baselineY, const char* text, double angle)
{
	return MD_Send_cd2td(
		MD_Command_FillText, 
		baselineX, baselineY, text, angle);
}

MD_Status MD_FontSize(double size)
{
	return MD_Send_cd(MD_Command_FontSize, size);
}

MD_Status MD_GlossyButton(
	double topLeftX, double topLeftY, double width, double height,
	double cornerRadius,
	unsigned int buttonColor, unsigned int borderColor)
{
	return MD_Send_cd5c2(
		MD_Command_GlossyButton,
		topLeftX, topLeftY, width, height, cornerRadius,
		buttonColor, borderColor);
}

MD_Status MD_GlossyButton1(
	double topLeftX, double topLeftY, double width, double height,
	double cornerRadius,
	unsigned int buttonColor, unsigned int borderColor,
	const char* text)
{
	return MD_Send_cd5c2t(
		MD_Command_GlossyButton,
		topLeftX, topLeftY, width, height, cornerRadius,
		buttonColor, borderColor,
		text);
}

MD_Status MD_Image(short x, short y, const char* filePath)
{
	return MD_Send_ci162t(MD_Command_Image, x, y, filePath);
}

MD_Status MD_Layer(unsigned char layerID)
{
	return MD_Send_cu8(MD_Command_Layer, layerID);
}

MD_Status MD_LineCap(MD_Cap lineCap)
{
	MD_StartFrame();
	
	MD_SendCommand(MD_Command_LineCap);
	
	MD_ArgLineCap(lineCap);
	
	MD_EndFrame();
	
	return MD_ProcessResponse();
}

MD_Status MD_LineJoin(MD_Join lineJoin)
{
	MD_StartFrame();
	
	MD_SendCommand(MD_Command_LineJoin);
	
	MD_ArgLineJoin(lineJoin);
	
	MD_EndFrame();
	
	return MD_ProcessResponse();
}

MD_Status MD_LineTo(double x, double y)
{
	return MD_Send_cd2(MD_Command_LineTo, x, y);
}

MD_Status MD_LineWidth(double width)
{
	return MD_Send_cd(MD_Command_LineWidth, width);
}

MD_Status MD_LinearGradient(
	double startX, double startY, unsigned int startColor, 
	double endX, double endY, unsigned int endColor)
{
	MD_StartFrame();
	
	MD_SendCommand(MD_Command_LinearGradient);
	
	MD_ArgReal(startX);
	MD_ArgReal(startY);
	MD_ArgColor(startColor);
	
	MD_ArgReal(endX);
	MD_ArgReal(endY);
	MD_ArgColor(endColor);
	
	MD_EndFrame();
	
	return MD_ProcessResponse();
}

MD_Status MD_LoadFont(const char* filePath)
{
	return MD_Send_ct(MD_Command_LoadFont, filePath);
}

MD_Status MD_LoadFont1(const char* filePath, unsigned char cacheInMemory)
{
	MD_StartFrame();
	
	MD_SendCommand(MD_Command_LoadFont);
	
	MD_ArgText(filePath);
	MD_ArgUInt8(cacheInMemory);
	
	MD_EndFrame();
	
	return MD_ProcessResponse();
}

MD_Status MD_MoveLayer(short topLeftX, short topLeftY)
{
	MD_StartFrame();
	
	MD_SendCommand(MD_Command_MoveLayer);
	
	MD_ArgInt16(topLeftX);
	MD_ArgInt16(topLeftY);
	
	MD_EndFrame();
	
	return MD_ProcessResponse();
}

MD_Status MD_MoveTo(double x, double y)
{
	return MD_Send_cd2(MD_Command_MoveTo, x, y);
}

MD_Status MD_RadialGradient(double startX, double startY, unsigned int startColor, double radius, unsigned int endColor)
{
	MD_StartFrame();
	
	MD_SendCommand(MD_Command_RadialGradient);
	
	MD_ArgReal(startX);
	MD_ArgReal(startY);
	MD_ArgColor(startColor);
	
	MD_ArgReal(radius);
	MD_ArgColor(endColor);
	
	MD_EndFrame();
	
	return MD_ProcessResponse();
}

MD_Status MD_Ping()
{
	return MD_Send_c(MD_Command_Ping);
}

MD_Status MD_RunScript(const char* filePath)
{
	return MD_Send_ct(MD_Command_RunScript, filePath);
}

MD_Status MD_Reset()
{
	return MD_Send_c(MD_Command_Reset);
}

MD_Status MD_SetPixel(short x, short y, const unsigned int* colorArray, unsigned int count)
{
	MD_StartFrame();
	
	MD_SendCommand(MD_Command_SetPixel);
	
	MD_ArgInt16(x);
	MD_ArgInt16(y);
	
	int i = 0;
	for(i = 0; i < count; i++)
	{
		MD_ArgColor(colorArray[i]);
	}
	
	MD_EndFrame();
	
	return MD_ProcessResponse();
}

MD_Status MD_UnloadFont()
{
	return MD_Send_c(MD_Command_UnloadFont);
}