/************************************************************
* M Display Library for the Moacon
* Michael Franklin
* 2013-09-09 - First Release
*
* To use simply add mdisplay.c and mdisplay.h to the Moacon
* Studio project nd #include "mdisplay.h" in your source file
*************************************************************/

#ifndef MD__H
#define MD__H

typedef enum MD_Status
{
	MD_Status_Success                = 0x0000,
	MD_Status_UnknownCommand         = 0x0001,
	MD_Status_WrongNumberOfArguments = 0x0002,
	MD_Status_InvalidArgument        = 0x0003,
	MD_Status_FileReadError          = 0x0004,
	MD_Status_MemoryAllocationError  = 0x0005,
	MD_Status_CommunicationError     = 0x0006,
	MD_Status_ChecksumError          = 0x0007,
	MD_Status_InvalidResponse        = 0xFFFD,
	MD_Status_Timeout                = 0xFFFE,
	MD_Status_UnspecifiedError       = 0xFFFF
} MD_Status;

typedef enum MD_Command
{
	MD_Command_ArcTo            = 0x00, 
	MD_Command_Backlight        = 0x01, 
	MD_Command_BevelButton      = 0x02, 
	MD_Command_Button           = 0x03, 
	MD_Command_Clear            = 0x04, 
	MD_Command_ClearPath        = 0x05, 
	// 0x06 reserved for ACK
	MD_Command_Color            = 0x07,
	MD_Command_CreateLayer      = 0x08,
	MD_Command_CurveTo          = 0x09,
	MD_Command_Delay            = 0x0A,
	MD_Command_DestroyLayer     = 0x0B,
	MD_Command_DisableFlush     = 0x0C,
	MD_Command_DrawArc          = 0x0D,
	MD_Command_DrawCircle       = 0x0E,
	MD_Command_DrawCurve        = 0x0F,
	MD_Command_DrawEllipse      = 0x10,
	MD_Command_DrawLine         = 0x11,
	MD_Command_DrawPath         = 0x12,
	MD_Command_DrawRectangle    = 0x13, 
	MD_Command_DrawText         = 0x14,
	// 0x15 Reserved for NAK
	MD_Command_FillCircle       = 0x16,
	MD_Command_FillEllipse      = 0x17,
	MD_Command_FillPath         = 0x18,
	MD_Command_FillRectangle    = 0x19,
	MD_Command_FillText         = 0x1A,
	MD_Command_FontSize         = 0x1B,
	MD_Command_GlossyButton     = 0x1C,
	MD_Command_Image            = 0x1D,
	MD_Command_Layer            = 0x1E,
	MD_Command_LineCap          = 0x1F,
	MD_Command_LineJoin         = 0x20,
	MD_Command_LineTo           = 0x21,
	MD_Command_LineWidth        = 0x22,
	MD_Command_LinearGradient   = 0x23,
	MD_Command_LoadFont         = 0x24,
	MD_Command_MoveLayer        = 0x25,
	MD_Command_MoveTo           = 0x26,
	MD_Command_Ping             = 0x27,
	MD_Command_RadialGradient   = 0x28,
	MD_Command_RunScript        = 0x29,
	MD_Command_Reset            = 0x2A,
	MD_Command_SetPixel         = 0x2C,
	MD_Command_UnloadFont       = 0x2D,
	MD_Command_ClosePath        = 0x32, 
	MD_Command_EnableFlush      = 0x33,
	MD_Command_DisableLayer     = 0x34,
	MD_Command_EnableLayer      = 0x35
} MD_Command;

typedef enum MD_Protocol
{
	MD_Protocol_UTF8   = 0,
	MD_Protocol_Binary = 1
} MD_Protocol;

typedef enum MD_Cap
{
	MD_Cap_Butt   = 0,
	MD_Cap_Round  = 1,
	MD_Cap_Square = 2
} MD_Cap;

typedef enum MD_Join
{
	MD_Join_Miter = 0,
	MD_Join_Round = 1,
	MD_Join_Bevel = 2 
} MD_Join;

/*************************************************************
* Opens the serial port for communication with the M Display
*************************************************************/
void MD_OpenCom(u8 comCh, u32 comBaud, u8 comMode);

/*************************************************************
* Sets the function to call in case of an error
*************************************************************/
void MD_SetOnError(void(*onError)(MD_Status status));

/*************************************************************
* Get the protocol being used for communication with the
* M Display
*************************************************************/
MD_Protocol MD_GetProtocol();

/*************************************************************
* Sets the protocol to use for communication with the
* M Display
*************************************************************/
void MD_SetProtocol(MD_Protocol protocol);

/*************************************************************
* Get whether or not to send/recieve a checksum
*************************************************************/
unsigned char MD_IsChecksumEnabled();

/*************************************************************
* Enables sending/receiving checksum
*************************************************************/
void MD_EnableChecksum();

/*************************************************************
* Disables sending/receiving checksum
*************************************************************/
void MD_DisableChecksum();

/*************************************************************
* Sets the amount of time in milliseconds to wait for a 
* response from the M Display
*************************************************************/
void MD_SetTimeout(unsigned short timeoutInMilliseconds);

/*************************************************************
* Gets the amount of time in milliseconds to wait for a 
* response from the M Display
*************************************************************/
unsigned short MD_GetTimeout();


/*************************************************************
* The following functions correspond to each of the individual
* command supported by the M Display.  See the M Display
* users's manual for more information.
*************************************************************/
MD_Status MD_ArcTo(double x, double y, double horizontalRadius);
MD_Status MD_ArcTo1(double x, double y, double horizontalRadius, double verticalRadius);
MD_Status MD_ArcTo2(double x, double y, double horizontalRadius, double verticalRadius, double rotationAngle);
MD_Status MD_ArcTo3(
	double x, double y, double horizontalRadius, double verticalRadius, double rotationAngle, 
	unsigned char largeArc);
MD_Status MD_ArcTo4(
	double x, double y, double horizontalRadius, double verticalRadius, double rotationAngle, 
	unsigned char largeArc, unsigned char sweep);

MD_Status MD_Backlight(unsigned char value);
MD_Status MD_BevelButton(
	double topLeftX, double topLeftY, double width, double height,
	unsigned int buttonColor, unsigned int topLeftColor, unsigned int bottomRightColor);
	
MD_Status MD_BevelButton1(
	double topLeftX, double topLeftY, double width, double height,
	unsigned int buttonColor, unsigned int topLeftColor, unsigned int bottomRightColor,
	const char* text);
	
MD_Status MD_Button(
	double topLeftX, double topLeftY, double width, double height, 
	double cornerRadius,
	unsigned int buttonColor, unsigned int borderColor);
	
MD_Status MD_Button1(
	double topLeftX, double topLeftY, double width, double height, 
	double cornerRadius,
	unsigned int buttonColor, unsigned int borderColor,
	const char* text);

MD_Status MD_Clear();
MD_Status MD_ClearPath();
MD_Status MD_ClosePath();
MD_Status MD_Color(unsigned int color);
MD_Status MD_CreateLayer(short topLeftX, short topLeftY, unsigned short width, unsigned short height, unsigned char layerID);
MD_Status MD_CurveTo(double x, double y, double control0X, double control0Y);
MD_Status MD_CurveTo1(double x, double y, double control0X, double control0Y, double control1X, double control1Y);

MD_Status MD_Delay(unsigned int milliseconds);
MD_Status MD_DestroyLayer(unsigned char layerID);
MD_Status MD_DisableFlush();
MD_Status MD_DisableLayer(unsigned char layerID);
MD_Status MD_DrawArc(
	double centerX, double centerY, 
	double horizontalRadius, 
	double startAngle, double sweep);
MD_Status MD_DrawArc1(
	double centerX, double centerY, 
	double horizontalRadius, double verticalRadius, 
	double startAngle, double sweep);
MD_Status MD_DrawCircle(double centerX, double centerY, double radius);
MD_Status MD_DrawCurve(
	double fromX, double fromY, 
	double control0X, double control0Y, 
	double toX, double toY);
MD_Status MD_DrawCurve1(
	double fromX, double fromY, 
	double control0X, double control0Y, 
	double control1X, double control1Y, 
	double toX, double toY);
MD_Status MD_DrawEllipse(double centerX, double centerY, double horizontalRadius, double verticalRadius);
MD_Status MD_DrawLine(double x0, double y0, double x1, double y1);
MD_Status MD_DrawPath();
MD_Status MD_DrawRectangle(double topLeftX, double topLeftY, double width, double height);
MD_Status MD_DrawRectangle1(
	double topLeftX, double topLeftY, double width, double height, 
	double cornerRadius);
MD_Status MD_DrawRectangle2(
	double topLeftX, double topLeftY, double width, double height, 
	double topLeftCornerRadius, double topRightCornerRadius, 
	double bottomLeftCornerRadius, double bottomRightCornerRadius);
MD_Status MD_DrawText(double baselineX, double baselineY, const char* text);
MD_Status MD_DrawText1(double baselineX, double baselineY, const char* text, double angle);
	
MD_Status MD_EnableFlush();
MD_Status MD_EnableLayer(unsigned char layerID);
	
MD_Status MD_FillCircle(double centerX, double centerY, double radius);
MD_Status MD_FillEllipse(double centerX, double centerY, double horizontalRadius, double verticalRadius);
MD_Status MD_FillPath();
MD_Status MD_FillRectangle(double topLeftX, double topLeftY, double width, double height);
MD_Status MD_FillRectangle1(
	double topLeftX, double topLeftY, double width, double height, 
	double cornerRadius);
MD_Status MD_FillRectangle2(
	double topLeftX, double topLeftY, double width, double height, 
	double topLeftCornerRadius, double topRightCornerRadius, 
	double bottomLeftCornerRadius, double bottomRightCornerRadius);
MD_Status MD_FillText(double baselineX, double baselineY, const char* text);
MD_Status MD_FillText1(double baselineX, double baselineY, const char* text, double angle);
MD_Status MD_FontSize(double size);

MD_Status MD_GlossyButton(
	double topLeftX, double topLeftY, double width, double height,
	double cornerRadius,
	unsigned int buttonColor, unsigned int borderColor);
MD_Status MD_GlossyButton1(
	double topLeftX, double topLeftY, double width, double height,
	double cornerRadius,
	unsigned int buttonColor, unsigned int borderColor,
	const char* text);
	
MD_Status MD_Image(short x, short y, const char* filePath);

MD_Status MD_Layer(unsigned char layerID);
MD_Status MD_LineCap(MD_Cap lineCap);
MD_Status MD_LineJoin(MD_Join lineJoin);
MD_Status MD_LineTo(double x, double y);
MD_Status MD_LineWidth(double width);
MD_Status MD_LinearGradient(
	double startX, double startY, unsigned int startColor, 
	double endX, double endY, unsigned int endColor);
MD_Status MD_LoadFont(const char* filePath);
MD_Status MD_LoadFont1(const char* filePath, unsigned char cacheInMemory);

MD_Status MD_MoveLayer(short topLeftX, short topLeftY);
MD_Status MD_MoveTo(double x, double y);

MD_Status MD_Ping();

MD_Status MD_RadialGradient(double startX, double startY, unsigned int startColor, double radius, unsigned int endColor);
MD_Status MD_RunScript(const char* filePath);
MD_Status MD_Reset();

MD_Status MD_SetPixel(short x, short y, const unsigned int* colorArray, unsigned int count);

MD_Status MD_UnloadFont();

#endif